/*
 *  C++ core and functions for BASIN.
 *  Copyright (C) 2006  BASIN team
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
/** \file ba_trait.h
    \brief Defines the mpi_type_trait information for use with ba

    These are the definitions of mpi_type_traits used for contiguous and non-contiguous datatypes.  Changes in ba soon to make unnecessary for user to change?*/
#ifndef BA_TRAIT_H
#define BA_TRAIT_H

#include "basin_base.h"
#include "basin_complex.h"
#include "Vector3.h"
#include "Vector2.h"
#include "gen_struct.h"


template<typename T> class mpi_type_trait;
template<typename T> class mpi_type_trait2;

//contiguous datatypes
template<>
class mpi_type_trait<int>{
 private:
  MPI_Datatype dataT;
 public:
  MPI_Datatype get(){
    MPI_Datatype xdt;
    xdt=MPI_INT;
    dataT=xdt;
    return dataT;
  };
};
template<>
class mpi_type_trait<unsigned int>{
 private:
  MPI_Datatype dataT;
 public:
  MPI_Datatype get(){
    MPI_Datatype xdt;
    xdt=MPI_INT;
    dataT=xdt;
    return dataT;
  };
};
template<>
class mpi_type_trait<char>{
 private:
  MPI_Datatype dataT;
 public:
  MPI_Datatype get(){
    MPI_Datatype xdt;
    xdt=MPI_CHAR;
    dataT=xdt;
    return dataT;
  };
};
template<>
class mpi_type_trait<float>{
 private:
  MPI_Datatype dataT;
 public:
  MPI_Datatype get(){ 
    MPI_Datatype xdt;
    xdt=MPI_FLOAT;
    dataT=xdt;
    return dataT;
  };
};
template<>
class mpi_type_trait<double>{
 private:
  MPI_Datatype dataT;
 public:
  MPI_Datatype get(){
    MPI_Datatype xdt;
    xdt=MPI_DOUBLE;
    dataT=xdt;
    return dataT;
  };
};
template<>
class mpi_type_trait<Vector3>{
 private:
  MPI_Datatype dataT;
 public:
  MPI_Datatype get(){
    MPI_Datatype xdt;
    MPI_Type_contiguous(3,MPI_DOUBLE,&xdt);
    MPI_Type_commit(&xdt);
    dataT=xdt;
    return dataT;
  }
};

template<>
class mpi_type_trait<Vector2>{
 private:
  MPI_Datatype dataT;
 public:
  MPI_Datatype get(){
    MPI_Datatype xdt;
    MPI_Type_contiguous(2,MPI_DOUBLE,&xdt);
    MPI_Type_commit(&xdt);
    dataT=xdt;
    return dataT;
  }
};

template<>
class mpi_type_trait<fftw_complex>{
 private:
  MPI_Datatype dataT;
 public:
  MPI_Datatype get(){
    MPI_Datatype xdt;
    MPI_Type_contiguous(2,MPI_DOUBLE,&xdt);
    MPI_Type_commit(&xdt);
    dataT=xdt;
    return dataT;
  }
};

//non-contiguous datatypes

template<>
class mpi_type_trait< non_c_d<double> >{
 private:
  MPI_Datatype dataT;
  //  static const MPI_Datatype da_ty=MPI_DOUBLE;
  //  static const int n1=3;
 public:
  //  MPI_Datatype get(int n_elem,int b_length,int stride,MPI_Datatype da_ty){
  //  MPI_Datatype get(int n1,int b_length,int stride){
  MPI_Datatype get(int n1,int stride){
    int b_length[2];
    MPI_Aint    displ[2], extent; 
    MPI_Datatype xdt,types[2];
    MPI_Type_extent(MPI_DOUBLE, &extent); 
    b_length[0]=n1;
    b_length[1]=1;
    displ[0]=0;
    displ[1]=stride*extent;
    types[0]=MPI_DOUBLE;
    types[1]=MPI_UB;
    //    MPI_Type_vector(n1,b_length,stride,MPI_DOUBLE,&xdt);
    MPI_Type_struct(2,b_length,displ,types,&xdt);
    MPI_Type_commit(&xdt);
    dataT=xdt;
    return dataT;
  }
};


template<>
class mpi_type_trait< non_c_d<char> >{
 private:
  MPI_Datatype dataT;
  //  static const MPI_Datatype da_ty=MPI_DOUBLE;
  //  static const int n1=3;
 public:
  //  MPI_Datatype get(int n_elem,int b_length,int stride,MPI_Datatype da_ty){
  //  MPI_Datatype get(int n1,int b_length,int stride){
  MPI_Datatype get(int n1,int stride){
    int b_length[2];
    MPI_Aint    displ[2], extent; 
    MPI_Datatype xdt,types[2];
    MPI_Type_extent(MPI_CHAR, &extent); 
    b_length[0]=n1;
    b_length[1]=1;
    displ[0]=0;
    displ[1]=stride*extent;
    types[0]=MPI_CHAR;
    types[1]=MPI_UB;
    //    MPI_Type_vector(n1,b_length,stride,MPI_DOUBLE,&xdt);
    MPI_Type_struct(2,b_length,displ,types,&xdt);
    MPI_Type_commit(&xdt);
    dataT=xdt;
    return dataT;
  }
};





template<>
class mpi_type_trait< non_c_d<float> >{
 private:
  MPI_Datatype dataT;
 public:
  MPI_Datatype get(int n1,int stride){
    int b_length[2];
    MPI_Aint    displ[2], extent; 
    MPI_Datatype xdt,types[2];
    MPI_Type_extent(MPI_FLOAT, &extent); 
    b_length[0]=n1;
    b_length[1]=1;
    displ[0]=0;
    displ[1]=stride*extent;
    types[0]=MPI_FLOAT;
    types[1]=MPI_UB;
    MPI_Type_struct(2,b_length,displ,types,&xdt);
    MPI_Type_commit(&xdt);
    dataT=xdt;
    return dataT;
  }
};




template<>
class mpi_type_trait< non_c_d<int> >{
 private:
  MPI_Datatype dataT;
 public:
  MPI_Datatype get(int n1,int stride){
    int b_length[2];
    MPI_Aint    displ[2], extent; 
    MPI_Datatype xdt,types[2];
    MPI_Type_extent(MPI_INT, &extent); 
    b_length[0]=n1;
    b_length[1]=1;
    displ[0]=0;
    displ[1]=stride*extent;
    types[0]=MPI_INT;
    types[1]=MPI_UB;
    MPI_Type_struct(2,b_length,displ,types,&xdt);
    MPI_Type_commit(&xdt);
    dataT=xdt;
    return dataT;
  }
};
template<>
class mpi_type_trait< non_c_d<unsigned int> >{
 private:
  MPI_Datatype dataT;
 public:
  MPI_Datatype get(int n1,int stride){
    int b_length[2];
    MPI_Aint    displ[2], extent; 
    MPI_Datatype xdt,types[2];
    MPI_Type_extent(MPI_INT, &extent); 
    b_length[0]=n1;
    b_length[1]=1;
    displ[0]=0;
    displ[1]=stride*extent;
    types[0]=MPI_INT;
    types[1]=MPI_UB;
    MPI_Type_struct(2,b_length,displ,types,&xdt);
    MPI_Type_commit(&xdt);
    dataT=xdt;
    return dataT;
  }
};
//=================================
template<>
class mpi_type_trait< non_c_d_md<double> >{
 private:
  MPI_Datatype dataT;
 public:
  MPI_Datatype get(std::vector<int> n1,std::vector<int> stride){
    /* n1 and stride have a dimension=(dimension_matrix-1)*/
    /* for A[i1,i2,i3,i4....iN]
       stride[0]=iN_max
       stride[1]=i(N-1)_max
       ....and so on
    */
    int number_dim;
    int ki;

    number_dim=stride.size();
    int tmp_length[2];
    MPI_Aint    tmp_displ[2], extent; 
    MPI_Datatype xdt[number_dim],types[2];
    MPI_Type_extent(MPI_DOUBLE, &extent); 
    types[0]=MPI_DOUBLE;
    types[1]=MPI_UB;
    tmp_length[0]=n1[0];
    tmp_length[1]=1;
    tmp_displ[0]=0;
    tmp_displ[1]=stride[0]*extent;
    //THIS LINE HERE CRASHES WITH LARGE NUMBER OF ELEMENTS!
    MPI_Type_struct(2,tmp_length,tmp_displ,types,&xdt[0]);
    MPI_Type_commit(&xdt[0]);
    ki=0;
    int tmp_stride=stride[0];
    for (int k=2;k<2*number_dim;k+=2)
      {
	types[0]=xdt[ki];
	types[1]=MPI_UB;
	ki+=1;
	tmp_stride*=stride[ki];
	tmp_length[0]=n1[ki];
	tmp_length[1]=1;
	tmp_displ[0]=0;
	tmp_displ[1]=tmp_stride*extent;
	MPI_Type_struct(2,tmp_length,tmp_displ,types,&xdt[ki]);
	MPI_Type_commit(&xdt[ki]);
      }
    dataT=xdt[ki];
    return dataT;
  }
};


template<>
class mpi_type_trait< non_c_d_md<char> >{
 private:
  MPI_Datatype dataT;
 public:
  MPI_Datatype get(std::vector<int> n1,std::vector<int> stride){
    /* n1 and stride have a dimension=(dimension_matrix-1)*/
    /* for A[i1,i2,i3,i4....iN]
       stride[0]=iN_max
       stride[1]=i(N-1)_max
       ....and so on
    */
    int number_dim;
    int ki;

    number_dim=stride.size();
    int tmp_length[2];
    MPI_Aint    tmp_displ[2], extent; 
    MPI_Datatype xdt[number_dim],types[2];
    MPI_Type_extent(MPI_CHAR, &extent); 
    types[0]=MPI_CHAR;
    types[1]=MPI_UB;
    tmp_length[0]=n1[0];
    tmp_length[1]=1;
    tmp_displ[0]=0;
    tmp_displ[1]=stride[0]*extent;
    //THIS LINE HERE CRASHES WITH LARGE NUMBER OF ELEMENTS!
    MPI_Type_struct(2,tmp_length,tmp_displ,types,&xdt[0]);
    MPI_Type_commit(&xdt[0]);
    ki=0;
    int tmp_stride=stride[0];
    for (int k=2;k<2*number_dim;k+=2)
      {
	types[0]=xdt[ki];
	types[1]=MPI_UB;
	ki+=1;
	tmp_stride*=stride[ki];
	tmp_length[0]=n1[ki];
	tmp_length[1]=1;
	tmp_displ[0]=0;
	tmp_displ[1]=tmp_stride*extent;
	MPI_Type_struct(2,tmp_length,tmp_displ,types,&xdt[ki]);
	MPI_Type_commit(&xdt[ki]);
      }
    dataT=xdt[ki];
    return dataT;
  }
};



template<>
class mpi_type_trait< non_c_d_md<float> >{
 private:
  MPI_Datatype dataT;
 public:
  MPI_Datatype get(std::vector<int> n1,std::vector<int> stride){
    /* n1 and stride have a dimension=(dimension_matrix-1)*/
    /* for A[i1,i2,i3,i4....iN]
       stride[0]=iN_max
       stride[1]=i(N-1)_max
       ....and so on
    */
    int number_dim;
    int ki;

    number_dim=stride.size();
    int tmp_length[2];
    MPI_Aint    tmp_displ[2], extent; 
    MPI_Datatype xdt[number_dim],types[2];

    MPI_Type_extent(MPI_FLOAT, &extent); 

    types[0]=MPI_FLOAT;
    types[1]=MPI_UB;
    tmp_length[0]=n1[0];
    tmp_length[1]=1;
    tmp_displ[0]=0;
    tmp_displ[1]=stride[0]*extent;

    MPI_Type_struct(2,tmp_length,tmp_displ,types,&xdt[0]);
    MPI_Type_commit(&xdt[0]);

    ki=0;
    int tmp_stride=stride[0];
    for (int k=2;k<2*number_dim;k+=2)
      {
	types[0]=xdt[ki];
	types[1]=MPI_UB;
	ki+=1;
	tmp_stride*=stride[ki];
	tmp_length[0]=n1[ki];
	tmp_length[1]=1;
	tmp_displ[0]=0;
	tmp_displ[1]=tmp_stride*extent;
	MPI_Type_struct(2,tmp_length,tmp_displ,types,&xdt[ki]);
	MPI_Type_commit(&xdt[ki]);
      }
    dataT=xdt[ki];
    return dataT;
  }
};
template<>
class mpi_type_trait< non_c_d_md<int> >{
 private:
  MPI_Datatype dataT;
 public:
  MPI_Datatype get(std::vector<int> n1,std::vector<int> stride){
    /* n1 and stride have a dimension=(dimension_matrix-1)*/
    /* for A[i1,i2,i3,i4....iN]
       stride[0]=iN_max
       stride[1]=i(N-1)_max
       ....and so on
    */
    int number_dim;
    int ki;

    number_dim=stride.size();
    int tmp_length[2];
    MPI_Aint    tmp_displ[2], extent; 
    MPI_Datatype xdt[number_dim],types[2];

    MPI_Type_extent(MPI_INT, &extent); 

    types[0]=MPI_INT;
    types[1]=MPI_UB;
    tmp_length[0]=n1[0];
    tmp_length[1]=1;
    tmp_displ[0]=0;
    tmp_displ[1]=stride[0]*extent;

    MPI_Type_struct(2,tmp_length,tmp_displ,types,&xdt[0]);
    MPI_Type_commit(&xdt[0]);

    ki=0;
    int tmp_stride=stride[0];
    for (int k=2;k<2*number_dim;k+=2)
      {
	types[0]=xdt[ki];
	types[1]=MPI_UB;
	ki+=1;
	tmp_stride*=stride[ki];
	tmp_length[0]=n1[ki];
	tmp_length[1]=1;
	tmp_displ[0]=0;
	tmp_displ[1]=tmp_stride*extent;
	MPI_Type_struct(2,tmp_length,tmp_displ,types,&xdt[ki]);
	MPI_Type_commit(&xdt[ki]);
      }
    dataT=xdt[ki];
    return dataT;
  }
};
template<>
class mpi_type_trait< non_c_d_md<unsigned int> >{
 private:
  MPI_Datatype dataT;
 public:
  MPI_Datatype get(std::vector<int> n1,std::vector<int> stride){
    /* n1 and stride have a dimension=(dimension_matrix-1)*/
    /* for A[i1,i2,i3,i4....iN]
       stride[0]=iN_max
       stride[1]=i(N-1)_max
       ....and so on
    */
    int number_dim;
    int ki;

    number_dim=stride.size();
    int tmp_length[2];
    MPI_Aint    tmp_displ[2], extent; 
    MPI_Datatype xdt[number_dim],types[2];

    MPI_Type_extent(MPI_INT, &extent); 

    types[0]=MPI_INT;
    types[1]=MPI_UB;
    tmp_length[0]=n1[0];
    tmp_length[1]=1;
    tmp_displ[0]=0;
    tmp_displ[1]=stride[0]*extent;

    MPI_Type_struct(2,tmp_length,tmp_displ,types,&xdt[0]);
    MPI_Type_commit(&xdt[0]);

    ki=0;
    int tmp_stride=stride[0];
    for (int k=2;k<2*number_dim;k+=2)
      {
	types[0]=xdt[ki];
	types[1]=MPI_UB;
	ki+=1;
	tmp_stride*=stride[ki];
	tmp_length[0]=n1[ki];
	tmp_length[1]=1;
	tmp_displ[0]=0;
	tmp_displ[1]=tmp_stride*extent;
	MPI_Type_struct(2,tmp_length,tmp_displ,types,&xdt[ki]);
	MPI_Type_commit(&xdt[ki]);
      }
    dataT=xdt[ki];
    return dataT;
  }
};
template<>
class mpi_type_trait< non_c_d_md<Vector3> >{
 private:
  MPI_Datatype dataT;
 public:
  MPI_Datatype get(std::vector<int> n1,std::vector<int> stride){
    /* n1 and stride have a dimension=(dimension_matrix-1)*/
    /* for A[i1,i2,i3,i4....iN]
       stride[0]=iN_max
       stride[1]=i(N-1)_max
       ....and so on
    */
    int number_dim;
    int ki;

    number_dim=stride.size();
    int tmp_length[2];
    MPI_Aint    tmp_displ[2], extent; 
    MPI_Datatype xdt[number_dim],types[2];
    mpi_type_trait<Vector3> local_mpt;
    MPI_Datatype local_datatype;
    local_datatype=local_mpt.get();
    MPI_Type_extent(local_datatype, &extent); 

    types[0]=local_datatype;
    types[1]=MPI_UB;
    tmp_length[0]=n1[0];
    tmp_length[1]=1;
    tmp_displ[0]=0;
    tmp_displ[1]=stride[0]*extent;

    MPI_Type_struct(2,tmp_length,tmp_displ,types,&xdt[0]);
    MPI_Type_commit(&xdt[0]);

    ki=0;
    int tmp_stride=stride[0];
    for (int k=2;k<2*number_dim;k+=2)
      {
	types[0]=xdt[ki];
	types[1]=MPI_UB;
	ki+=1;
	tmp_stride*=stride[ki];
	tmp_length[0]=n1[ki];
	tmp_length[1]=1;
	tmp_displ[0]=0;
	tmp_displ[1]=tmp_stride*extent;
	MPI_Type_struct(2,tmp_length,tmp_displ,types,&xdt[ki]);
	MPI_Type_commit(&xdt[ki]);
      }
    dataT=xdt[ki];
    return dataT;
  }
};
template<>
class mpi_type_trait< non_c_d_md<Vector2> >{
 private:
  MPI_Datatype dataT;
 public:
  MPI_Datatype get(std::vector<int> n1,std::vector<int> stride){
    /* n1 and stride have a dimension=(dimension_matrix-1)*/
    /* for A[i1,i2,i3,i4....iN]
       stride[0]=iN_max
       stride[1]=i(N-1)_max
       ....and so on
    */
    int number_dim;
    int ki;

    number_dim=stride.size();
    int tmp_length[2];
    MPI_Aint    tmp_displ[2], extent; 
    MPI_Datatype xdt[number_dim],types[2];
    mpi_type_trait<Vector2> local_mpt;
    MPI_Datatype local_datatype;
    local_datatype=local_mpt.get();
    MPI_Type_extent(local_datatype, &extent); 

    types[0]=local_datatype;
    types[1]=MPI_UB;
    tmp_length[0]=n1[0];
    tmp_length[1]=1;
    tmp_displ[0]=0;
    tmp_displ[1]=stride[0]*extent;

    MPI_Type_struct(2,tmp_length,tmp_displ,types,&xdt[0]);
    MPI_Type_commit(&xdt[0]);

    ki=0;
    int tmp_stride=stride[0];
    for (int k=2;k<2*number_dim;k+=2)
      {
	types[0]=xdt[ki];
	types[1]=MPI_UB;
	ki+=1;
	tmp_stride*=stride[ki];
	tmp_length[0]=n1[ki];
	tmp_length[1]=1;
	tmp_displ[0]=0;
	tmp_displ[1]=tmp_stride*extent;
	MPI_Type_struct(2,tmp_length,tmp_displ,types,&xdt[ki]);
	MPI_Type_commit(&xdt[ki]);
      }
    dataT=xdt[ki];
    return dataT;
  }
};
template<>
class mpi_type_trait< non_c_d_md<fftw_complex> >{
 private:
  MPI_Datatype dataT;
 public:
  MPI_Datatype get(std::vector<int> n1,std::vector<int> stride){
    /* n1 and stride have a dimension=(dimension_matrix-1)*/
    /* for A[i1,i2,i3,i4....iN]
       stride[0]=iN_max
       stride[1]=i(N-1)_max
       ....and so on
    */
    int number_dim;
    int ki;

    number_dim=stride.size();
    int tmp_length[2];
    MPI_Aint    tmp_displ[2], extent; 
    MPI_Datatype xdt[number_dim],types[2];
    mpi_type_trait<fftw_complex> local_mpt;
    MPI_Datatype local_datatype;
    local_datatype=local_mpt.get();
    MPI_Type_extent(local_datatype, &extent); 

    types[0]=local_datatype;
    types[1]=MPI_UB;
    tmp_length[0]=n1[0];
    tmp_length[1]=1;
    tmp_displ[0]=0;
    tmp_displ[1]=stride[0]*extent;

    MPI_Type_struct(2,tmp_length,tmp_displ,types,&xdt[0]);
    MPI_Type_commit(&xdt[0]);

    ki=0;
    int tmp_stride=stride[0];
    for (int k=2;k<2*number_dim;k+=2)
      {
	types[0]=xdt[ki];
	types[1]=MPI_UB;
	ki+=1;
	tmp_stride*=stride[ki];
	tmp_length[0]=n1[ki];
	tmp_length[1]=1;
	tmp_displ[0]=0;
	tmp_displ[1]=tmp_stride*extent;
	MPI_Type_struct(2,tmp_length,tmp_displ,types,&xdt[ki]);
	MPI_Type_commit(&xdt[ki]);
      }
    dataT=xdt[ki];
    return dataT;
  }
};

template<typename T>
class mpi_type_trait2<aux_vect<T> >{
 private:
  MPI_Datatype dataT;
  MPI_Datatype local_dataT;
 public:

  
  MPI_Datatype get(){
    mpi_type_trait<T> mpt;
    local_dataT=mpt.get();
    MPI_Datatype xdt;
    int count=2;
    int blockcounts[2];
    MPI_Aint extent;
    MPI_Aint offsets[2];
    MPI_Datatype oldtypes[2];
    MPI_Type_extent(local_dataT,&extent);
    //MPI_Type_extent(MPI_INT,&extent1);
    oldtypes[0]=local_dataT;
    offsets[0]=0;
    blockcounts[0]=1;
    oldtypes[1]=MPI_INT;
    offsets[1]=extent;
    blockcounts[1]=1;
    MPI_Type_struct(count,blockcounts,offsets,oldtypes,&xdt);
    MPI_Type_commit(&xdt);
    dataT=xdt;
    return dataT;
  }
};


#endif
