/*
 *  C++ core and functions for BASIN.
 *  Copyright (C) 2006  BASIN team
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#ifndef basin_h
#define basin_h

#include "Region.h"
#include "Data.h"
#include "Attribute.h"

/*
GLOBAL VARIABLES ARE BAD IN GENERAL BUT in this case they 
are needed to define some convenient things for the users
like bout defined as 'if BASIN_ID==0 cout'
BASIN_ID is the id of a process
BASIN_NPROCS is the number of processes
*/
extern int BASIN_ID, BASIN_NPROCS;

/* BASIN_COMM is the BASIN communicator. 
To be used all over the code instead of MPI_COMM_WORLD
*/
extern MPI_Comm BASIN_COMM;
/* 
Some useful functions to initialize/finalize MPI
Everything below will eventually go into a separate header file
*/

///Initialize MPI; It returns a vector<int> with two elements the first is the number of elements, the second is the process id. These values are also put in the two global variables BASIN_ID and BASIN_NPROCS
vector<int> basin_initialize(int &ID=BASIN_ID, int &NP=BASIN_NPROCS,MPI_Comm c=MPI_COMM_WORLD);

/// Calls MPI_Finalize
void basin_finalize();

#define bout if (BASIN_ID==0) cout
#define berr if (BASIN_ID==0) cerr

#endif
