#ifndef FFTW_BASIN_H
#define FFTW_BASIN_H

#include "basin_base.h"

using namespace std;


inline fftw_complex operator+(fftw_complex a, double b){
  fftw_complex c=a;
  c.re += b;
  return c;
}

inline fftw_complex operator-(fftw_complex a, double b){
  fftw_complex c=a;
  c.re -= b;
  return c;
}
inline fftw_complex operator*(fftw_complex a, double b){
  fftw_complex c=a;
  c.re *=b;
  c.im *=b;
  return c;
}
inline fftw_complex operator/(fftw_complex a, double b){
  fftw_complex c=a;
  c.re /= b;
  c.im /= b;
  return c;
}


inline fftw_complex operator+(fftw_complex a, fftw_complex b)
{
  fftw_complex c;
  c.im = a.im + b.im;
  c.re = a.re + b.re;
  return c;
}
inline fftw_complex operator-(fftw_complex a, fftw_complex b){
  fftw_complex c;
  c.im = a.im - b.im;
  c.re = a.re - b.re;
  return c;
}
inline fftw_complex operator*(fftw_complex a, fftw_complex b){
  fftw_complex c;
  c.re = a.re*b.re - a.im*b.im;
  c.im = a.re*b.im + a.im*b.re;
  return c;
}
inline fftw_complex operator/(fftw_complex a, fftw_complex b){
  fftw_complex c=b;
  c.im *= -1;
  fftw_complex top = a*c;
  fftw_complex bottom = b*c;
  return top / bottom.re;
}


inline fftw_complex operator+(fftw_complex a, float b){
  fftw_complex c=a;
  c.re += b;
  return c;
}
inline fftw_complex operator-(fftw_complex a, float b){
  fftw_complex c=a;
  c.re -= b;
  return c;
}
inline fftw_complex operator*(fftw_complex a, float b){
  fftw_complex c=a;
  c.re *=b;
  c.im *=b;
  return c;
}
inline fftw_complex operator/(fftw_complex a, float b){
  fftw_complex c=a;
  c.re /= b;
  c.im /= b;
  return c;
}


inline fftw_complex operator+(fftw_complex a, int b){
  fftw_complex c=a;
  c.re += b;
  return c;
}
inline fftw_complex operator-(fftw_complex a, int b){
  fftw_complex c=a;
  c.re -= b;
  return c;
}
inline fftw_complex operator*(fftw_complex a, int b){
  fftw_complex c=a;
  c.re *=b;
  c.im *=b;
  return c;
}
inline fftw_complex operator/(fftw_complex a, int b){
  fftw_complex c=a;
  c.re /= b;
  c.im /= b;
  return c;
}








inline fftw_complex operator+(double a, fftw_complex b){
  fftw_complex c;
  c.re = a;
  c.im = 0.0;
  return c+b;
}
inline fftw_complex operator-(double a, fftw_complex b){
  fftw_complex c;
  c.re = a;
  c.im = 0.0;
  return c+b;
}
inline fftw_complex operator*(double a, fftw_complex b){
  fftw_complex c;
  c.re = a;
  c.im = 0.0;
  return c+b;
}
inline fftw_complex operator/(double a, fftw_complex b){
  fftw_complex c;
  c.re = a;
  c.im = 0.0;
  return c+b;
}


inline fftw_complex operator+(float a, fftw_complex b){
  fftw_complex c;
  c.re = a;
  c.im = 0.0;
  return c+b;
}
inline fftw_complex operator-(float a, fftw_complex b){
  fftw_complex c;
  c.re = a;
  c.im = 0.0;
  return c+b;
}
inline fftw_complex operator*(float a, fftw_complex b){
  fftw_complex c;
  c.re = a;
  c.im = 0.0;
  return c+b;
}
inline fftw_complex operator/(float a, fftw_complex b){
  fftw_complex c;
  c.re = a;
  c.im = 0.0;
  return c+b;
}


inline fftw_complex operator+(int a, fftw_complex b){
  fftw_complex c;
  c.re = a;
  c.im = 0.0;
  return c+b;
}
inline fftw_complex operator-(int a, fftw_complex b){
  fftw_complex c;
  c.re = a;
  c.im = 0.0;
  return c+b;
}
inline fftw_complex operator*(int a, fftw_complex b){
  fftw_complex c;
  c.re = a;
  c.im = 0.0;
  return c+b;
}
inline fftw_complex operator/(int a, fftw_complex b){
  fftw_complex c;
  c.re = a;
  c.im = 0.0;
  return c+b;
}




inline fftw_complex pow(fftw_complex a, fftw_complex b){
  /*
    (a+bi)^(c+di) = [(a^2 + b^2)^(c/2)][e^(-d*atan(b/a))][cos(c*atan(b/a) + 0.5*d*ln(a^2 + b^2)) + i*sin(c*atan(b/a) + 0.5*d*ln(a^2 + b^2))]
    is the formula
    a=a.re
    b=a.im
    c=b.re
    d=b.im
  */
  double part1 = pow(pow(a.re,2) + pow(a.im,2),b.re/2);
  double part2 = exp(-b.im*atan(a.im/a.re));
  double part3real = cos(b.re*atan(a.im/a.re) + 0.5 * log(pow(a.re,2) + pow(a.im,2)));
  double part3imag = sin(b.re*atan(a.im/a.re) + 0.5 * log(pow(a.re,2) + pow(a.im,2)));
  fftw_complex c;
  c.re = part1*part2*part3real;
  c.im = part1*part2*part3imag;
  return c;
}
inline fftw_complex pow(fftw_complex a, double b){
  fftw_complex c;
  c.im = 0;
  c.re = b;
  return pow(a,c);
}

inline double abs(fftw_complex a){
  return sqrt(pow(a.re,2)+pow(a.im,2));
}
inline  double fabs(fftw_complex a){
  return sqrt(pow(a.re,2)+pow(a.im,2));
}
inline fftw_complex log(fftw_complex a){
  /*
    formula:
    log(a+bi) = log(abs(a+bi)) + i*atan(b/a)
    a=a.re
    b=a.im
  */
  fftw_complex b;
  b.re = log(fabs(a));
  b.im = atan(a.im/a.re);
  return b;
}
inline fftw_complex log10(fftw_complex a){
  /*
    logb(x) = log(x)/log(b)
  */
  return log(a)/log(10.0);
}

inline fftw_complex pow(double a, fftw_complex b){
  fftw_complex c;
  c.im =0;
  c.re = a;
  return pow(c,b);
}

inline fftw_complex sqrt(fftw_complex a){return pow(a,0.5);}
inline fftw_complex exp(fftw_complex a){return pow(exp(1.0),a);}
inline fftw_complex exp10(fftw_complex a){return pow(10.0,a);}

inline double real(fftw_complex b){
  return b.re;
}

inline double imag(fftw_complex b){
  return b.im;
}

inline ostream &operator << (ostream &out_file, const fftw_complex& y)
{
  out_file<<" "<<y.re<<"+i*"<<y.im<<" "<<endl;
  return out_file;
}

/*
THESE FUNCTIONS ARE NOT DEFINED:
-if you try to call one of these functions on an Attribute of fftw_comple
 BA will take care of letting the user know that the operation is not defined 
 at run time

-if you try to call one of these function on an Attribute_element 
 the compiler will issue an error

fftw_complex sin(fftw_complex a){return a;}
fftw_complex cos(fftw_complex a){return a;}
fftw_complex cosh(fftw_complex a){return a;}
fftw_complex sinh(fftw_complex a){return a;}
fftw_complex acos(fftw_complex a){return a;}
fftw_complex asin(fftw_complex a){return a;}
fftw_complex acosh(fftw_complex a){return a;}
fftw_complex asinh(fftw_complex a){return a;}
fftw_complex tan(fftw_complex a){return a;}
fftw_complex atan(fftw_complex a){return a;}

*/

#endif
