/* 
   functionS.C included in BA.C

   This file contains all the specializations needed for BA math functions.
   The default behavior for most functions is to return an error message 
   saying that the operation is not defined. For those datatypes for
   which math operations are defined this is the file where the
   definition should be implemented. The templatized math functions
   defined in BA.C will take care of all other types.

   Enrico Vesperini

*/

//EXP FUNCTION
template <>  inline
ba<double> exp(ba<double> &a)
{
  ba<double> local_ba;
  int d_v=a.ba_size();
  vector<int> dims;
  dims=a.ba_get_dimensions();
  MPI_Comm cm=a.ba_get_comm();
  local_ba.ba_create(dims,cm,0);
  int loc_d_v=a.ba_get_loc_dimv();
  double loc_g;
  for (int i=0;i<(loc_d_v)*(d_v/dims[0]);i++)
    {
      loc_g=a.ba_get_access(i);
      double temp=exp(loc_g);
      local_ba.ba_set(i,temp);
    }
  return local_ba;
}


template <>  inline
ba<float> exp(ba<float> &a)
{
  ba<float> local_ba;
  int d_v=a.ba_size();
  vector<int> dims;
  dims=a.ba_get_dimensions();
  MPI_Comm cm=a.ba_get_comm();
  local_ba.ba_create(dims,cm,0);
  int loc_d_v=a.ba_get_loc_dimv();
  double loc_g;
  for (int i=0;i<(loc_d_v)*(d_v/dims[0]);i++)
    {
      loc_g=a.ba_get_access(i);
      float temp=exp(loc_g);
      local_ba.ba_set(i,temp);
    }
  return local_ba;
}


template <>  inline
ba<fftw_complex> exp(ba<fftw_complex> &a)
{
  ba<fftw_complex> local_ba;
  int d_v=a.ba_size();
  vector<int> dims;
  dims=a.ba_get_dimensions();
  MPI_Comm cm=a.ba_get_comm();
  local_ba.ba_create(dims,cm,0);
  int loc_d_v=a.ba_get_loc_dimv();
  fftw_complex loc_g;
  for (int i=0;i<(loc_d_v)*(d_v/dims[0]);i++)
    {
      loc_g=a.ba_get_access(i);
      fftw_complex temp=exp(loc_g);
      local_ba.ba_set(i,temp);
    }
  return local_ba;
}


//POW FUNCTION

template <> inline
ba<double> pow(ba<double> &a,double z)
{
  ba<double> local_ba;
  int d_v=a.ba_size();
  vector<int> dims;
  dims=a.ba_get_dimensions();
  MPI_Comm cm=a.ba_get_comm();
  local_ba.ba_create(dims,cm,0);
  int loc_d_v=a.ba_get_loc_dimv();
  double loc_g;
  for (int i=0;i<(loc_d_v)*(d_v/dims[0]);i++)
    {
      loc_g=a.ba_get_access(i);
      double temp=pow(loc_g,z);
      local_ba.ba_set(i,temp);
    }
  return local_ba;
}

template <> inline
ba<float> pow(ba<float> &a,double z)
{
  ba<float> local_ba;
  int d_v=a.ba_size();
  vector<int> dims;
  dims=a.ba_get_dimensions();
  MPI_Comm cm=a.ba_get_comm();
  local_ba.ba_create(dims,cm,0);
  int loc_d_v=a.ba_get_loc_dimv();
  float loc_g;
  for (int i=0;i<(loc_d_v)*(d_v/dims[0]);i++)
    {
      loc_g=a.ba_get_access(i);
      float temp=(float)pow((double)loc_g,z);
      local_ba.ba_set(i,temp);
    }
  return local_ba;
}


template <> inline
ba<fftw_complex> pow(ba<fftw_complex> &a,double z)
{
  ba<fftw_complex> local_ba;
  int d_v=a.ba_size();
  vector<int> dims;
  dims=a.ba_get_dimensions();
  MPI_Comm cm=a.ba_get_comm();
  local_ba.ba_create(dims,cm,0);
  int loc_d_v=a.ba_get_loc_dimv();
  fftw_complex loc_g;
  loc_g.re=0.0;
  loc_g.im=0.0;
  for (int i=0;i<(loc_d_v)*(d_v/dims[0]);i++)
    {
      loc_g=a.ba_get_access(i);
      fftw_complex temp=pow(loc_g,z);
      local_ba.ba_set(i,temp);
    }
  return local_ba;
}

template <> inline
ba<fftw_complex> pow(ba<double> &a,fftw_complex z)
{
  ba<fftw_complex> local_ba;
  int d_v=a.ba_size();
  vector<int> dims;
  dims=a.ba_get_dimensions();
  MPI_Comm cm=a.ba_get_comm();
  local_ba.ba_create(dims,cm,0);
  int loc_d_v=a.ba_get_loc_dimv();
  double loc_g;
  for (int i=0;i<(loc_d_v)*(d_v/dims[0]);i++)
    {
      loc_g=a.ba_get_access(i);
      fftw_complex temp=pow(loc_g,z);
      local_ba.ba_set(i,temp);
    }
  return local_ba;
}




// LOG FUNCTION

template <> inline
ba<double> log(ba<double> &a)
{
  ba<double> local_ba;
  int d_v=a.ba_size();
  vector<int> dims;
  dims=a.ba_get_dimensions();
  MPI_Comm cm=a.ba_get_comm();
  local_ba.ba_create(dims,cm,0);
  int loc_d_v=a.ba_get_loc_dimv();
  double loc_g;
  for (int i=0;i<(loc_d_v)*(d_v/dims[0]);i++)
    {
      loc_g=a.ba_get_access(i);
      double temp=log(loc_g);
      local_ba.ba_set(i,temp);
    }
  return local_ba;
}


template <> inline
ba<float> log(ba<float> &a)
{
  ba<float> local_ba;
  int d_v=a.ba_size();
  vector<int> dims;
  dims=a.ba_get_dimensions();
  MPI_Comm cm=a.ba_get_comm();
  local_ba.ba_create(dims,cm,0);
  int loc_d_v=a.ba_get_loc_dimv();
  float loc_g;
  for (int i=0;i<(loc_d_v)*(d_v/dims[0]);i++)
    {
      loc_g=a.ba_get_access(i);
      float temp=logf(loc_g);
      local_ba.ba_set(i,temp);
    }
  return local_ba;
}

template <> inline
ba<fftw_complex> log(ba<fftw_complex> &a)
{
  ba<fftw_complex> local_ba;
  int d_v=a.ba_size();
  vector<int> dims;
  dims=a.ba_get_dimensions();
  MPI_Comm cm=a.ba_get_comm();
  local_ba.ba_create(dims,cm,0);
  int loc_d_v=a.ba_get_loc_dimv();
  fftw_complex loc_g;
  loc_g.re=0.0;
  loc_g.im=0.0;
  for (int i=0;i<(loc_d_v)*(d_v/dims[0]);i++)
    {
      loc_g=a.ba_get_access(i);
      fftw_complex temp=log(loc_g);
      local_ba.ba_set(i,temp);
    }
  return local_ba;
}


// LOG10 FUNCTION

template <> inline
ba<double> log10(ba<double> &a)
{
  ba<double> local_ba;
  int d_v=a.ba_size();
  vector<int> dims;
  dims=a.ba_get_dimensions();
  MPI_Comm cm=a.ba_get_comm();
  local_ba.ba_create(dims,cm,0);
  int loc_d_v=a.ba_get_loc_dimv();
  double loc_g;
  for (int i=0;i<(loc_d_v)*(d_v/dims[0]);i++)
    {
      loc_g=a.ba_get_access(i);
      double temp=log10(loc_g);
      local_ba.ba_set(i,temp);
    }
  return local_ba;
}


template <> inline
ba<float> log10(ba<float> &a)
{
  ba<float> local_ba;
  int d_v=a.ba_size();
  vector<int> dims;
  dims=a.ba_get_dimensions();
  MPI_Comm cm=a.ba_get_comm();
  local_ba.ba_create(dims,cm,0);
  int loc_d_v=a.ba_get_loc_dimv();
  float loc_g;
  for (int i=0;i<(loc_d_v)*(d_v/dims[0]);i++)
    {
      loc_g=a.ba_get_access(i);
      float temp=log10f(loc_g);
      local_ba.ba_set(i,temp);
    }
  return local_ba;
}

template <> inline
ba<fftw_complex> log10(ba<fftw_complex> &a)
{
  ba<fftw_complex> local_ba;
  int d_v=a.ba_size();
  vector<int> dims;
  dims=a.ba_get_dimensions();
  MPI_Comm cm=a.ba_get_comm();
  local_ba.ba_create(dims,cm,0);
  int loc_d_v=a.ba_get_loc_dimv();
  fftw_complex loc_g;
  loc_g.re=0.0;
  loc_g.im=0.0;
  for (int i=0;i<(loc_d_v)*(d_v/dims[0]);i++)
    {
      loc_g=a.ba_get_access(i);
      fftw_complex temp=log10(loc_g);
      local_ba.ba_set(i,temp);
    }
  return local_ba;
}



// SIN FUNCTION

template <> inline
ba<double> sin(ba<double> &a)
{
  ba<double> local_ba;
  int d_v=a.ba_size();
  vector<int> dims;
  dims=a.ba_get_dimensions();
  MPI_Comm cm=a.ba_get_comm();
  local_ba.ba_create(dims,cm,0);
  int loc_d_v=a.ba_get_loc_dimv();
  double loc_g;
  for (int i=0;i<(loc_d_v)*(d_v/dims[0]);i++)
    {
      loc_g=a.ba_get_access(i);
      double temp=sin(loc_g);
      local_ba.ba_set(i,temp);
    }
  return local_ba;
}


template <> inline
ba<float> sin(ba<float> &a)
{
  ba<float> local_ba;
  int d_v=a.ba_size();
  vector<int> dims;
  dims=a.ba_get_dimensions();
  MPI_Comm cm=a.ba_get_comm();
  local_ba.ba_create(dims,cm,0);
  int loc_d_v=a.ba_get_loc_dimv();
  float loc_g;
  for (int i=0;i<(loc_d_v)*(d_v/dims[0]);i++)
    {
      loc_g=a.ba_get_access(i);
      float temp=sinf(loc_g);
      local_ba.ba_set(i,temp);
    }
  return local_ba;
}


template <> inline
ba<double> cos(ba<double> &a)
{
  ba<double> local_ba;
  int d_v=a.ba_size();
  vector<int> dims;
  dims=a.ba_get_dimensions();
  MPI_Comm cm=a.ba_get_comm();
  local_ba.ba_create(dims,cm,0);
  int loc_d_v=a.ba_get_loc_dimv();
  double loc_g;
  for (int i=0;i<(loc_d_v)*(d_v/dims[0]);i++)
    {
      loc_g=a.ba_get_access(i);
      double temp=cos(loc_g);
      local_ba.ba_set(i,temp);
      }
  return local_ba;
}



template <> inline
ba<float> cos(ba<float> &a)
{
  ba<float> local_ba;
  int d_v=a.ba_size();
  vector<int> dims;
  dims=a.ba_get_dimensions();
  MPI_Comm cm=a.ba_get_comm();
  local_ba.ba_create(dims,cm,0);
  int loc_d_v=a.ba_get_loc_dimv();
  float loc_g;
  for (int i=0;i<(loc_d_v)*(d_v/dims[0]);i++)
    {
      loc_g=a.ba_get_access(i);
      float temp=cosf(loc_g);
      local_ba.ba_set(i,temp);
      }
  return local_ba;
}



template <> inline
ba<double> cosh(ba<double> &a)
{
  ba<double> local_ba;
  int d_v=a.ba_size();
  vector<int> dims;
  dims=a.ba_get_dimensions();
  MPI_Comm cm=a.ba_get_comm();
  local_ba.ba_create(dims,cm,0);
  int loc_d_v=a.ba_get_loc_dimv();
  double loc_g;
  for (int i=0;i<(loc_d_v)*(d_v/dims[0]);i++)
    {
      loc_g=a.ba_get_access(i);
      double temp=cosh(loc_g);
      local_ba.ba_set(i,temp);
      }
  return local_ba;
}

template <> inline
ba<float> cosh(ba<float> &a)
{
  ba<float> local_ba;
  int d_v=a.ba_size();
  vector<int> dims;
  dims=a.ba_get_dimensions();
  MPI_Comm cm=a.ba_get_comm();
  local_ba.ba_create(dims,cm,0);
  int loc_d_v=a.ba_get_loc_dimv();
  float loc_g;
  for (int i=0;i<(loc_d_v)*(d_v/dims[0]);i++)
    {
      loc_g=a.ba_get_access(i);
      float temp=coshf(loc_g);
      local_ba.ba_set(i,temp);
      }
  return local_ba;
}

template <> inline
ba<double> sinh(ba<double> &a)
{
  ba<double> local_ba;
  int d_v=a.ba_size();
  vector<int> dims;
  dims=a.ba_get_dimensions();
  MPI_Comm cm=a.ba_get_comm();
  local_ba.ba_create(dims,cm,0);
  int loc_d_v=a.ba_get_loc_dimv();
  double loc_g;
  for (int i=0;i<(loc_d_v)*(d_v/dims[0]);i++)
    {
      loc_g=a.ba_get_access(i);
      double temp=sinh(loc_g);
      local_ba.ba_set(i,temp);
      }
  return local_ba;
}

template <> inline
ba<float> sinh(ba<float> &a)
{
  ba<float> local_ba;
  int d_v=a.ba_size();
  vector<int> dims;
  dims=a.ba_get_dimensions();
  MPI_Comm cm=a.ba_get_comm();
  local_ba.ba_create(dims,cm,0);
  int loc_d_v=a.ba_get_loc_dimv();
  float loc_g;
  for (int i=0;i<(loc_d_v)*(d_v/dims[0]);i++)
    {
      loc_g=a.ba_get_access(i);
      float temp=sinhf(loc_g);
      local_ba.ba_set(i,temp);
      }
  return local_ba;
}

template <> inline
ba<double> asin(ba<double> &a)
{
  ba<double> local_ba;
  int d_v=a.ba_size();
  vector<int> dims;
  dims=a.ba_get_dimensions();
  MPI_Comm cm=a.ba_get_comm();
  local_ba.ba_create(dims,cm,0);
  int loc_d_v=a.ba_get_loc_dimv();
  double loc_g;
  for (int i=0;i<(loc_d_v)*(d_v/dims[0]);i++)
    {
      loc_g=a.ba_get_access(i);
      double temp=asin(loc_g);
      local_ba.ba_set(i,temp);
      }
  return local_ba;
}

template <> inline
ba<float> asin(ba<float> &a)
{
  ba<float> local_ba;
  int d_v=a.ba_size();
  vector<int> dims;
  dims=a.ba_get_dimensions();
  MPI_Comm cm=a.ba_get_comm();
  local_ba.ba_create(dims,cm,0);
  int loc_d_v=a.ba_get_loc_dimv();
  float loc_g;
  for (int i=0;i<(loc_d_v)*(d_v/dims[0]);i++)
    {
      loc_g=a.ba_get_access(i);
      float temp=asinf(loc_g);
      local_ba.ba_set(i,temp);
      }
  return local_ba;
}

template <> inline
ba<double> acos(ba<double> &a)
{
  ba<double> local_ba;
  int d_v=a.ba_size();
  vector<int> dims;
  dims=a.ba_get_dimensions();
  MPI_Comm cm=a.ba_get_comm();
  local_ba.ba_create(dims,cm,0);
  int loc_d_v=a.ba_get_loc_dimv();
  double loc_g;
  for (int i=0;i<(loc_d_v)*(d_v/dims[0]);i++)
    {
      loc_g=a.ba_get_access(i);
      double temp=acos(loc_g);
      local_ba.ba_set(i,temp);
      }
  return local_ba;
}

template <> inline
ba<float> acos(ba<float> &a)
{
  ba<float> local_ba;
  int d_v=a.ba_size();
  vector<int> dims;
  dims=a.ba_get_dimensions();
  MPI_Comm cm=a.ba_get_comm();
  local_ba.ba_create(dims,cm,0);
  int loc_d_v=a.ba_get_loc_dimv();
  float loc_g;
  for (int i=0;i<(loc_d_v)*(d_v/dims[0]);i++)
    {
      loc_g=a.ba_get_access(i);
      float temp=acosf(loc_g);
      local_ba.ba_set(i,temp);
      }
  return local_ba;
}


template <> inline
ba<double> asinh(ba<double> &a)
{
  ba<double> local_ba;
  int d_v=a.ba_size();
  vector<int> dims;
  dims=a.ba_get_dimensions();
  MPI_Comm cm=a.ba_get_comm();
  local_ba.ba_create(dims,cm,0);
  int loc_d_v=a.ba_get_loc_dimv();
  double loc_g;
  for (int i=0;i<(loc_d_v)*(d_v/dims[0]);i++)
    {
      loc_g=a.ba_get_access(i);
      double temp=asinh(loc_g);
      local_ba.ba_set(i,temp);
      }
  return local_ba;
}


template <> inline
ba<float> asinh(ba<float> &a)
{
  ba<float> local_ba;
  int d_v=a.ba_size();
  vector<int> dims;
  dims=a.ba_get_dimensions();
  MPI_Comm cm=a.ba_get_comm();
  local_ba.ba_create(dims,cm,0);
  int loc_d_v=a.ba_get_loc_dimv();
  float loc_g;
  for (int i=0;i<(loc_d_v)*(d_v/dims[0]);i++)
    {
      loc_g=a.ba_get_access(i);
      float temp=asinhf(loc_g);
      local_ba.ba_set(i,temp);
      }
  return local_ba;
}


template <> inline
ba<double> acosh(ba<double> &a)
{
  ba<double> local_ba;
  int d_v=a.ba_size();
  vector<int> dims;
  dims=a.ba_get_dimensions();
  MPI_Comm cm=a.ba_get_comm();
  local_ba.ba_create(dims,cm,0);
  int loc_d_v=a.ba_get_loc_dimv();
  double loc_g;
  for (int i=0;i<(loc_d_v)*(d_v/dims[0]);i++)
    {
      loc_g=a.ba_get_access(i);
      double temp=acosh(loc_g);
      local_ba.ba_set(i,temp);
      }
  return local_ba;
}

template <> inline
ba<float> acosh(ba<float> &a)
{
  ba<float> local_ba;
  int d_v=a.ba_size();
  vector<int> dims;
  dims=a.ba_get_dimensions();
  MPI_Comm cm=a.ba_get_comm();
  local_ba.ba_create(dims,cm,0);
  int loc_d_v=a.ba_get_loc_dimv();
  float loc_g;
  for (int i=0;i<(loc_d_v)*(d_v/dims[0]);i++)
    {
      loc_g=a.ba_get_access(i);
      float temp=acoshf(loc_g);
      local_ba.ba_set(i,temp);
      }
  return local_ba;
}

template <> inline
ba<double> abs(ba<double> &a)
{
  ba<double> local_ba;
  int d_v=a.ba_size();
  vector<int> dims;
  dims=a.ba_get_dimensions();
  MPI_Comm cm=a.ba_get_comm();
  local_ba.ba_create(dims,cm,0);
  int loc_d_v=a.ba_get_loc_dimv();
  double loc_g;
  for (int i=0;i<(loc_d_v)*(d_v/dims[0]);i++)
    {
      loc_g=a.ba_get_access(i);
      double temp=fabs(loc_g);
      local_ba.ba_set(i,temp);
      }
  return local_ba;
}

template <> inline
ba<float> abs(ba<float> &a)
{
  ba<float> local_ba;
  int d_v=a.ba_size();
  vector<int> dims;
  dims=a.ba_get_dimensions();
  MPI_Comm cm=a.ba_get_comm();
  local_ba.ba_create(dims,cm,0);
  int loc_d_v=a.ba_get_loc_dimv();
  float loc_g;
  for (int i=0;i<(loc_d_v)*(d_v/dims[0]);i++)
    {
      loc_g=a.ba_get_access(i);
      float temp=fabsf(loc_g);
      local_ba.ba_set(i,temp);
      }
  return local_ba;
}


template <> inline
ba<int> abs(ba<int> &a)
{
  ba<int> local_ba;
  int d_v=a.ba_size();
  vector<int> dims;
  dims=a.ba_get_dimensions();
  MPI_Comm cm=a.ba_get_comm();
  local_ba.ba_create(dims,cm,0);
  int loc_d_v=a.ba_get_loc_dimv();
  int loc_g;
  for (int i=0;i<(loc_d_v)*(d_v/dims[0]);i++)
    {
      loc_g=a.ba_get_access(i);
      int temp=abs(loc_g);
      local_ba.ba_set(i,temp);
      }
  return local_ba;
}


template <> inline
ba<double> sqrt(ba<double> &a)
{
  ba<double> local_ba;
  int d_v=a.ba_size();
  vector<int> dims;
  dims=a.ba_get_dimensions();
  MPI_Comm cm=a.ba_get_comm();
  local_ba.ba_create(dims,cm,0);
  int loc_d_v=a.ba_get_loc_dimv();
  double loc_g;
  for (int i=0;i<(loc_d_v)*(d_v/dims[0]);i++)
    {
      loc_g=a.ba_get_access(i);
      double temp=sqrt(loc_g);
      local_ba.ba_set(i,temp);
      }
  return local_ba;
}

template <> inline
ba<float> sqrt(ba<float> &a)
{
  ba<float> local_ba;
  int d_v=a.ba_size();
  vector<int> dims;
  dims=a.ba_get_dimensions();
  MPI_Comm cm=a.ba_get_comm();
  local_ba.ba_create(dims,cm,0);
  int loc_d_v=a.ba_get_loc_dimv();
  float loc_g;
  for (int i=0;i<(loc_d_v)*(d_v/dims[0]);i++)
    {
      loc_g=a.ba_get_access(i);
      float temp=sqrtf(loc_g);
      local_ba.ba_set(i,temp);
      }
  return local_ba;
}

template <> inline
ba<fftw_complex> sqrt(ba<fftw_complex> &a)
{
  ba<fftw_complex> local_ba;
  int d_v=a.ba_size();
  vector<int> dims;
  dims=a.ba_get_dimensions();
  MPI_Comm cm=a.ba_get_comm();
  local_ba.ba_create(dims,cm,0);
  int loc_d_v=a.ba_get_loc_dimv();
  fftw_complex loc_g;
  loc_g.re=0.0;
  loc_g.im=0.0;
  for (int i=0;i<(loc_d_v)*(d_v/dims[0]);i++)
    {
      loc_g=a.ba_get_access(i);
      fftw_complex temp=sqrt(loc_g);
      local_ba.ba_set(i,temp);
      }
  return local_ba;
}


template <> inline
ba<double> exp10(ba<double> &a)
{
  ba<double> local_ba;
  int d_v=a.ba_size();
  vector<int> dims;
  dims=a.ba_get_dimensions();
  MPI_Comm cm=a.ba_get_comm();
  local_ba.ba_create(dims,cm,0);
  int loc_d_v=a.ba_get_loc_dimv();
  double loc_g;
  for (int i=0;i<(loc_d_v)*(d_v/dims[0]);i++)
    {
      loc_g=a.ba_get_access(i);
      double temp=pow(10.0,loc_g);
      local_ba.ba_set(i,temp);
      }
  return local_ba;
}

template <> inline
ba<float> exp10(ba<float> &a)
{
  ba<float> local_ba;
  int d_v=a.ba_size();
  vector<int> dims;
  dims=a.ba_get_dimensions();
  MPI_Comm cm=a.ba_get_comm();
  local_ba.ba_create(dims,cm,0);
  int loc_d_v=a.ba_get_loc_dimv();
  float loc_g;
  for (int i=0;i<(loc_d_v)*(d_v/dims[0]);i++)
    {
      loc_g=a.ba_get_access(i);
      float temp=powf(10.0,loc_g);
      local_ba.ba_set(i,temp);
      }
  return local_ba;
}

template <> inline
ba<fftw_complex> exp10(ba<fftw_complex> &a)
{
  ba<fftw_complex> local_ba;
  int d_v=a.ba_size();
  vector<int> dims;
  dims=a.ba_get_dimensions();
  MPI_Comm cm=a.ba_get_comm();
  local_ba.ba_create(dims,cm,0);
  int loc_d_v=a.ba_get_loc_dimv();
  fftw_complex loc_g;
  for (int i=0;i<(loc_d_v)*(d_v/dims[0]);i++)
    {
      loc_g=a.ba_get_access(i);
      fftw_complex temp=pow(10.0,loc_g);
      local_ba.ba_set(i,temp);
      }
  return local_ba;
}

template <> inline
ba<double> atan(ba<double> &a)
{
  ba<double> local_ba;
  int d_v=a.ba_size();
  vector<int> dims;
  dims=a.ba_get_dimensions();
  MPI_Comm cm=a.ba_get_comm();
  local_ba.ba_create(dims,cm,0);
  int loc_d_v=a.ba_get_loc_dimv();
  double loc_g;
  for (int i=0;i<(loc_d_v)*(d_v/dims[0]);i++)
    {
      loc_g=a.ba_get_access(i);
      double temp=atan(loc_g);
      local_ba.ba_set(i,temp);
      }
  return local_ba;
}
template <> inline
ba<float> atan(ba<float> &a)
{
  ba<float> local_ba;
  int d_v=a.ba_size();
  vector<int> dims;
  dims=a.ba_get_dimensions();
  MPI_Comm cm=a.ba_get_comm();
  local_ba.ba_create(dims,cm,0);
  int loc_d_v=a.ba_get_loc_dimv();
  float loc_g;
  for (int i=0;i<(loc_d_v)*(d_v/dims[0]);i++)
    {
      loc_g=a.ba_get_access(i);
      float temp=atanf(loc_g);
      local_ba.ba_set(i,temp);
      }
  return local_ba;
}


inline ba<double> real(ba<fftw_complex> &a)
{
  ba<double> local_ba;
  int d_v=a.ba_size();
  vector<int> dims;
  dims=a.ba_get_dimensions();
  MPI_Comm cm=a.ba_get_comm();
  local_ba.ba_create(dims,cm,0);
  int loc_d_v=a.ba_get_loc_dimv();
  fftw_complex loc_g;
  for (int i=0;i<(loc_d_v)*(d_v/dims[0]);i++)
    {
      loc_g=a.ba_get_access(i);
      double temp=loc_g.re;
      local_ba.ba_set(i,temp);
      }
  return local_ba;
}

inline ba<double> real(ba<double> &a)
{
  ba<double> local_ba;
  int d_v=a.ba_size();
  vector<int> dims;
  dims=a.ba_get_dimensions();
  MPI_Comm cm=a.ba_get_comm();
  local_ba.ba_create(dims,cm,0);
  int loc_d_v=a.ba_get_loc_dimv();
  double loc_g;
  for (int i=0;i<(loc_d_v)*(d_v/dims[0]);i++)
    {
      loc_g=a.ba_get_access(i);
      double temp=loc_g;
      local_ba.ba_set(i,temp);
      }
  return local_ba;
}

inline ba<float> real(ba<float> &a)
{
  ba<float> local_ba;
  int d_v=a.ba_size();
  vector<int> dims;
  dims=a.ba_get_dimensions();
  MPI_Comm cm=a.ba_get_comm();
  local_ba.ba_create(dims,cm,0);
  int loc_d_v=a.ba_get_loc_dimv();
  float loc_g;
  for (int i=0;i<(loc_d_v)*(d_v/dims[0]);i++)
    {
      loc_g=a.ba_get_access(i);
      float temp=loc_g;
      local_ba.ba_set(i,temp);
      }
  return local_ba;
}

inline ba<int> real(ba<int> &a)
{
  ba<int> local_ba;
  int d_v=a.ba_size();
  vector<int> dims;
  dims=a.ba_get_dimensions();
  MPI_Comm cm=a.ba_get_comm();
  local_ba.ba_create(dims,cm,0);
  int loc_d_v=a.ba_get_loc_dimv();
  int loc_g;
  for (int i=0;i<(loc_d_v)*(d_v/dims[0]);i++)
    {
      loc_g=a.ba_get_access(i);
      int temp=loc_g;
      local_ba.ba_set(i,temp);
      }
  return local_ba;
}




inline ba<double> imag(ba<fftw_complex> &a)
{
  ba<double> local_ba;
  int d_v=a.ba_size();
  vector<int> dims;
  dims=a.ba_get_dimensions();
  MPI_Comm cm=a.ba_get_comm();
  local_ba.ba_create(dims,cm,0);
  int loc_d_v=a.ba_get_loc_dimv();
  fftw_complex loc_g;
  for (int i=0;i<(loc_d_v)*(d_v/dims[0]);i++)
    {
      loc_g=a.ba_get_access(i);
      double temp=loc_g.im;
      local_ba.ba_set(i,temp);
      }
  return local_ba;
}

inline ba<double> imag(ba<double> &a)
{
  ba<double> local_ba;
  int d_v=a.ba_size();
  vector<int> dims;
  dims=a.ba_get_dimensions();
  MPI_Comm cm=a.ba_get_comm();
  local_ba.ba_create(dims,cm,0);
  int loc_d_v=a.ba_get_loc_dimv();
  double loc_g;
  for (int i=0;i<(loc_d_v)*(d_v/dims[0]);i++)
    {
      double temp=0.0;
      local_ba.ba_set(i,temp);
      }
  return local_ba;
}

inline ba<float> imag(ba<float> &a)
{
  ba<float> local_ba;
  int d_v=a.ba_size();
  vector<int> dims;
  dims=a.ba_get_dimensions();
  MPI_Comm cm=a.ba_get_comm();
  local_ba.ba_create(dims,cm,0);
  int loc_d_v=a.ba_get_loc_dimv();
  float loc_g;
  for (int i=0;i<(loc_d_v)*(d_v/dims[0]);i++)
    {
      float temp=0.0;
      local_ba.ba_set(i,temp);
      }
  return local_ba;
}

inline ba<int> imag(ba<int> &a)
{
  ba<int> local_ba;
  int d_v=a.ba_size();
  vector<int> dims;
  dims=a.ba_get_dimensions();
  MPI_Comm cm=a.ba_get_comm();
  local_ba.ba_create(dims,cm,0);
  int loc_d_v=a.ba_get_loc_dimv();
  int loc_g;
  for (int i=0;i<(loc_d_v)*(d_v/dims[0]);i++)
    {
      int temp=0;
      local_ba.ba_set(i,temp);
      }
  return local_ba;
}







template <> inline
ba<double> ba_Sort_BA(ba<double> &a)
{
   int start=0;
  int length=0;
  int myrank;
  int numb_proc;
  MPI_Status status;
  vector<aux_vect<double> > temp;
  temp=a.make_sortable();
  a.ba_Sort(temp);
  ba<double> local_ba;
  int d_v=a.ba_size();
  vector<int> dims;
  dims=a.ba_get_dimensions();

  MPI_Comm cm=a.ba_get_comm();
  MPI_Comm_rank(cm, &myrank );
  MPI_Comm_size(cm,&numb_proc);

  local_ba.ba_create(dims,cm,0);
  double* values;
  values=new double[int(temp.size())];
  for (int k=0;k<(int(temp.size()));k++)
    {
      values[k]=(temp[k].val);
    }
  length=temp.size();
  if (myrank!=0)
    {
      MPI_Send(&length,1,MPI_INT,0,0,cm);
      MPI_Send(values,temp.size(),MPI_DOUBLE,0,0,cm);
    }
  else
    {
      int remote_length=0;
      vector<int> llo;
      llo.push_back(0);
      vector<int> hhi;
      hhi.push_back(length-1);

      local_ba.ba_put(llo,hhi,values);
            start+=temp.size();
      for (int i_rank=1;i_rank<numb_proc;i_rank++)
	{
	  MPI_Recv(&remote_length,1,MPI_INT,i_rank,0,cm,&status);
	  double* remote_values;
	  remote_values=new double[remote_length];
	  MPI_Recv(remote_values,remote_length,MPI_DOUBLE,i_rank,0,cm,&status);
	  llo[0]=start;
	  hhi[0]=(start+remote_length-1);
	  local_ba.ba_put(llo,hhi,remote_values);
	  start+=remote_length;
	  delete[] remote_values;
	  }
    }
  delete[] values;
  return local_ba;
}


template <> inline
ba<float> ba_Sort_BA(ba<float> &a)
{
   int start=0;
  int length=0;
  int myrank;
  int numb_proc;
  MPI_Status status;
  vector<aux_vect<float> > temp;
  temp=a.make_sortable();
  a.ba_Sort(temp);
  ba<float> local_ba;
  int d_v=a.ba_size();
  vector<int> dims;
  dims=a.ba_get_dimensions();

  MPI_Comm cm=a.ba_get_comm();
  MPI_Comm_rank( cm, &myrank );
  MPI_Comm_size(cm,&numb_proc);

  local_ba.ba_create(dims,cm,0);
  float* values;
  values=new float[int(temp.size())];
  for (int k=0;k<(int(temp.size()));k++)
    {
      values[k]=(temp[k].val);
    }
  length=temp.size();
  if (myrank!=0)
    {
      MPI_Send(&length,1,MPI_INT,0,0,cm);
      MPI_Send(values,temp.size(),MPI_FLOAT,0,0,cm);
    }
  else
    {
      int remote_length=0;
      vector<int> llo;
      llo.push_back(0);
      vector<int> hhi;
      hhi.push_back(length-1);

      local_ba.ba_put(llo,hhi,values);
            start+=temp.size();
      for (int i_rank=1;i_rank<numb_proc;i_rank++)
	{
	  MPI_Recv(&remote_length,1,MPI_INT,i_rank,0,cm,&status);
	  float* remote_values;
	  remote_values=new float[remote_length];
	  MPI_Recv(remote_values,remote_length,MPI_FLOAT,i_rank,0,cm,&status);
	  llo[0]=start;
	  hhi[0]=(start+remote_length-1);
	  local_ba.ba_put(llo,hhi,remote_values);
	  start+=remote_length;
	  delete[] remote_values;
	  }
    }
  delete[] values;
  return local_ba;
}


template <> inline
ba<int> ba_Sort_BA(ba<int> &a)
{
   int start=0;
  int length=0;
  int myrank;
  int numb_proc;
  MPI_Status status;
  vector<aux_vect<int> > temp;
  temp=a.make_sortable();
  a.ba_Sort(temp);
  ba<int> local_ba;
  int d_v=a.ba_size();
  vector<int> dims;
  dims=a.ba_get_dimensions();

  MPI_Comm cm=a.ba_get_comm();
  MPI_Comm_rank( cm, &myrank );
  MPI_Comm_size(cm,&numb_proc);

  local_ba.ba_create(dims,cm,0);
  int* values;
  values=new int[int(temp.size())];
  for (int k=0;k<(int(temp.size()));k++)
    {
      values[k]=(temp[k].val);
    }
  length=temp.size();
  if (myrank!=0)
    {
      MPI_Send(&length,1,MPI_INT,0,0,cm);
      MPI_Send(values,temp.size(),MPI_INT,0,0,cm);
    }
  else
    {
      int remote_length=0;
      vector<int> llo;
      llo.push_back(0);
      vector<int> hhi;
      hhi.push_back(length-1);

      local_ba.ba_put(llo,hhi,values);
            start+=temp.size();
      for (int i_rank=1;i_rank<numb_proc;i_rank++)
	{
	  MPI_Recv(&remote_length,1,MPI_INT,i_rank,0,cm,&status);
	  int* remote_values;
	  remote_values=new int[remote_length];
	  MPI_Recv(remote_values,remote_length,MPI_INT,i_rank,0,cm,&status);
	  llo[0]=start;
	  hhi[0]=(start+remote_length-1);
	  local_ba.ba_put(llo,hhi,remote_values);
	  start+=remote_length;
	  delete[] remote_values;
	  }
    }
  delete[] values;
  return local_ba;
}



template <> inline
ba<int> index_sorted(ba<double> &a)
{
   int start=0;
  int length=0;
  int myrank;
  int numb_proc;
  MPI_Status status;
  vector<aux_vect<double> > temp;
  temp=a.make_sortable();
  a.ba_Sort(temp);
  ba<int> local_ba;
  int d_v=a.ba_size();
  vector<int> dims;
  dims=a.ba_get_dimensions();

  MPI_Comm cm=a.ba_get_comm();
  MPI_Comm_rank(cm, &myrank );
  MPI_Comm_size(cm,&numb_proc);

  local_ba.ba_create(dims,cm,0);
  int* values;
  values=new int[int(temp.size())];
  for (int k=0;k<(int(temp.size()));k++)
    {
      values[k]=(temp[k].ind);
    }
  length=temp.size();
  if (myrank!=0)
    {
      MPI_Send(&length,1,MPI_INT,0,0,cm);
      MPI_Send(values,temp.size(),MPI_INT,0,0,cm);
    }
  else
    {
      int remote_length=0;
      vector<int> llo;
      llo.push_back(0);
      vector<int> hhi;
      hhi.push_back(length-1);

      local_ba.ba_put(llo,hhi,values);
            start+=temp.size();
      for (int i_rank=1;i_rank<numb_proc;i_rank++)
	{
	  MPI_Recv(&remote_length,1,MPI_INT,i_rank,0,cm,&status);
	  int* remote_values;
	  remote_values=new int[remote_length];
	  MPI_Recv(remote_values,remote_length,MPI_INT,i_rank,0,cm,&status);
	  llo[0]=start;
	  hhi[0]=(start+remote_length-1);
	  local_ba.ba_put(llo,hhi,remote_values);
	  start+=remote_length;
	  delete[] remote_values;
	  }
    }
  delete[] values;
  return local_ba;
}

template <> inline
double sum_BA(ba<double> &a)
{
  double local_ba=0;
  int d_v=a.ba_size();
  vector<int> dims;
  dims=a.ba_get_dimensions();
  MPI_Comm cm=a.ba_get_comm();
  int loc_d_v=a.ba_get_loc_dimv();
  double loc_g=0;
  for (int i=0;i<(loc_d_v)*(d_v/dims[0]);i++)
    {
      loc_g+=a.ba_get_access(i);
    }
  MPI_Allreduce(&loc_g,&local_ba,1,MPI_DOUBLE,MPI_SUM,cm);
  return local_ba;
}

template <> inline
float sum_BA(ba<float> &a)
{
  float local_ba=0;
  int d_v=a.ba_size();
  vector<int> dims;
  dims=a.ba_get_dimensions();
  MPI_Comm cm=a.ba_get_comm();
  int loc_d_v=a.ba_get_loc_dimv();
  float loc_g=0;
  for (int i=0;i<(loc_d_v)*(d_v/dims[0]);i++)
    {
      loc_g+=a.ba_get_access(i);
    }
  MPI_Allreduce(&loc_g,&local_ba,1,MPI_FLOAT,MPI_SUM,cm);
  return local_ba;
}


template <> inline
int sum_BA(ba<int> &a)
{
  int local_ba=0;
  int d_v=a.ba_size();
  vector<int> dims;
  dims=a.ba_get_dimensions();
  MPI_Comm cm=a.ba_get_comm();
  int loc_d_v=a.ba_get_loc_dimv();
  int loc_g=0;
  for (int i=0;i<(loc_d_v)*(d_v/dims[0]);i++)
    {
      loc_g+=a.ba_get_access(i);
    }
  MPI_Allreduce(&loc_g,&local_ba,1,MPI_INT,MPI_SUM,cm);
  return local_ba;
}




template <> inline
fftw_complex sum_BA(ba<fftw_complex> &a)
{
  MPI_Datatype data_type;
  mpi_type_trait<fftw_complex> mpt;
  data_type=mpt.get();
  fftw_complex local_ba;
  local_ba.re=0.0;
  local_ba.im=0.0;
  int d_v=a.ba_size();
  vector<int> dims;
  dims=a.ba_get_dimensions();
  MPI_Comm cm=a.ba_get_comm();
  int loc_d_v=a.ba_get_loc_dimv();
  fftw_complex loc_g;
  loc_g.re=0.0;;
  loc_g.im=0.0;;
  for (int i=0;i<(loc_d_v)*(d_v/dims[0]);i++)
    {
      loc_g=loc_g+a.ba_get_access(i);
    }
  MPI_Allreduce(&loc_g,&local_ba,1,data_type,MPI_SUM,cm);
  return local_ba;
}









//userf still to be done



