#ifndef __VISIT_MODULE_H__
#define __VISIT_MODULE_H__ 

#include "basin_package.h"


// All of the mesh_name arguments are optional now
// Some functions don't have a mesh_name argument at all now

/**
 * Expose an Attribute on a 3 dimensional point mesh. The first 3 Attributes define the mesh 
 * while the 4th Attribute contains the data to be displayed.
 */
void visit_expose_discrete(Attribute& x, Attribute& y, Attribute& z, Attribute& scalar, string visit_name);

/*
 * Exposes the variable and then tells Visit to generate a pseudocolor plot of the data.
 */
void visit_plot_discrete(Attribute& x, Attribute& y, Attribute& z, Attribute& scalar, string visit_name);

/**
 * Expose an Attribute on a 2 dimensional point mesh. The first 2 Attributes define the mesh
 * while the 3rd Attribute contains the data to be displayed.
 */
void visit_expose_discrete(Attribute& x, Attribute& y, Attribute& scalar, string visit_name); 

/*
 * Exposes the variable and then tells Visit to generate a pseudocolor plot of the data.
 */
void visit_plot_discrete(Attribute& x, Attribute& y, Attribute& scalar, string visit_name); 

/**
 * Expose an Attribute on a 3 dimensional point mesh. The first argument is a List object, the
 * next three strings are names of the x, y, and z dimesion Attributes. The 4th string is
 * the variable to plot. The last two strings are the name of the Mesh and Variable to create
 * in VisIt.
 */
void visit_expose_discrete(List& l, string x, string y, string z, string scalar, string visit_name);

/*
 * Exposes the variable and then tells Visit to generate a pseudocolor plot of the data.
 */
void visit_plot_discrete(List& l, string x, string y, string z, string scalar, string visit_name);

/**
 * Plot an Attribute on a 2 dimensional point mesh. The first argument is a List object, the
 * next two strings are names of the x and y dimesion Attributes. The 3rd string is
 * the variable to plot. The last two strings are the name of the Mesh and Variable to create
 * in VisIt.
 */
void visit_expose_discrete(List& l, string x, string y, string scalar, string visit_name);

/*
 * Exposes the variable and then tells Visit to generate a pseudocolor plot of the data.
 */
void visit_plot_discrete(List& l, string x, string y, string scalar, string visit_name);

/**
 * Plot an Attribute on a rectilinear mesh. The vectors ll, ur, and nx define
 * the mesh and the Attribute contains the data to be displayed.
 */
void visit_expose_continuous(vector<double> ll, vector<double> ur, vector<int> nx, Attribute &a, string visit_name);

/**
 * Exposes the variable and then tells Visit to generate a volume plot of the data.
 */
void visit_plot_continuous(vector<double> ll, vector<double> ur, vector<int> nx, Attribute &a, string visit_name);

/**
 * Plot an Attribute from a Grid on a rectilinear mesh. The Context of the Grid
 * will define the mesh. The string attr_name should be an Attribute belonging
 * to the Grid.
 */
void visit_expose_continuous(Grid& g, string attr_name, string visit_name);

/**
 * Exposes the variable and then tells Visit to generate a volume plot of the data.
 */
void visit_plot_continuous(Grid& g, string attr_name, string visit_name);

/**
 * Plot a free Attribute using the context from an existing Grid. A rectilinear mesh is formed from
 * the context.
 */
void visit_expose_continuous(Grid& g, Attribute& attr, string visit_name);

/**
 * Exposes the variable and then tells Visit to generate a volume plot of the data.
 */
void visit_plot_continuous(Grid& g, Attribute& attr, string visit_name);


/**
 * Creates an instance of the visit_plot package. Calling this will create a thread
 * that responds to commands from VisIt. This must be called before any of the plot
 * commands are used.
 */
void visit_module_start();

/**
 * Blocks the caller until the VisIt plot package thread exits. This is useful if
 * simuations are complete but the connection to VisIt should remain open for
 * further plotting.
 */
void visit_module_join();


/*
 * These are not needed by users of the package.
 */
void visit_loop();

int prep_vis_connection();


#endif //__VISIT_MODULE_H__
