# File:     basin_cluster.py

# Purpose:  This file defines an interface to the IPython controlled 
#           cluster. It defines an adapter to their interface in order
#           to present a simplified version of the cluster controls.
#           It also provides some conveinence functions to automate
#           connecting to the cluster in addition to other tasks.

#TODO   Rename this file to something else

#TODO   Look into a "whose" (inspired by Matlab) 
#       type function that shows the variables in the
#       user's namespace on the cluster. This should be similar to the 
#       dir() function, but filtered in some way. This can probably build
#       upon the ipython1 RemoteController.keys() method

from IPython.ColorANSI import TermColors

import ipython1.kernel.api as api



class BasinCluster:

    def __init__(self, host = ''):
        self.cluster = None
        self.engine0 = None

        if host != '':
            print "BasinCluster connecting to host: " + host
            self.connect(host)

        # These are the targets that %autopx sends commands
        # to. We want commands to go to all nodes
        self.magicTargets = 'all'

    def _isConnectedDecorator(func):
        def connectedTestDec(*args):
            self = args[0]
            if self.cluster != None:
                return func(*args)
            else:
                print ("%s Error: You are not connected to the cluster! Please use the connect method before continuing" % TermColors.Red)
        return connectedTestDec
    
    def _formatFeedback(self, ret):
        # Removes all of the output except for the first
        # element, which is from node zero and all that we
        # care about by default
        pass
        #del ret[1:len(ret)]
        return ret

    def _printFeedback(self, ret):
        # Expects a dictionary 'ret' with all of the feedback from
        # the IPython engines.
        # We only want to print feedback from Node zero
        
        if isinstance(ret, list):
            for f in ret:
                if f['id'] == 0:
                    print "%s BASIN input[%i]:%s %s" % \
                        (TermColors.Blue, f['commandIndex'],
                        TermColors.Normal, f['stdin'])
                    if f['stdout']:
                        print "%s BASIN output[%i]:%s %s" % \
                            (TermColors.Green, f['commandIndex'],
                            TermColors.Normal, f['stdout'])
                    if f['stderr']:
                        print "%s BASIN error[%i]:%s %s" % \
                            (TermColors.Red ,f['commandIndex'], 
                            TermColors.Normal, f['stderr'])
                    
                    # no sense in processing the rest of the returned
                    # data, since we only want feedback from node 0
                    break



    def connect(self, host):
        """Connect to the cluster on <host> that BASIN is running on.
        """
        # Find out what port IPython is using
        port = api.defaultRemoteController[1]
        # Get the IPython interface object
        if self.cluster == None:
            self.cluster = api.RemoteController((host, port))
        # Make sure the interface is connected properly
        # self.cluster.connect()
        # Make an alias to the cluster's engine 0
        #self.engine0 = self.cluster[0]
        # Make this controller the active one by default
        try:
            __IPYTHON__.activeController = self
        except NameError:
            print "Magic commands only work when used within the IPython client. %autopx will not work!"
        
        # try to import the basin library
        self.cluster.executeAll("from basin import *")

    def setBlock(self, block):
        """Set either True or False to tell the controller to block
        after each command is issued"""
        self.cluster.block = block
    
    @_isConnectedDecorator    
    def execute(self, targets, lines):
        """ Run the Python code provided on the cluster.
        """
        
        ret = self.cluster.execute(targets, lines)

        # ret now contains the return values for all clusters
        return self._formatFeedback(ret)

    @_isConnectedDecorator
    def iexecuteAll(self, lines):
        # When autopx is enabled, this method gets called
        # so it isn't useful directly, but must be defined for this mode
        self.execute(lines)

    @_isConnectedDecorator
    def getIDs(self):
        """Return a list of active engine ids"""
        return (self.cluster.getIDs())

    @_isConnectedDecorator
    def queueStatus(self):
        return self.cluster.queueStatus(0)

    @_isConnectedDecorator
    def clearQueue(self):
        self.cluster.clearQueueAll()

    @_isConnectedDecorator
    def killCluster(self):
        self.cluster.killAll(controller=True)

